// Bugs-B-Gone for C/C++
//
// Version 1.4 (Saturday, 23 September 2000)
//
// Created by Matthijs Hollemans 
// mahlzeit@bigfoot.com
// http://www.weirdexperiments.com/
//
// Bugs-B-Gone is public domain. No warranties expressed or implied.
//
// This package offers a few simple macros that allow you to write debugging 
// output to stdout. Nothing special, but it works for me. For more "internal
// debugging" power, you may wish to look at Fred Fish's DBUG package.
//
// Typically, you would put something like this at the top of your main 
// source (or include) file:
// 
//   #define  BUGS_B_GONE
//   #define  BUGS_B_GONE_PREFIX "MyApp"
//   #include "Bugs-B-Gone.h"
//
// The various symbols have the following meaning:
//
//     BUGS_B_GONE
//
//         -- Enables or disables the Bugs-B-Gone package.
//
//     BUGS_B_GONE_PREFIX
//
//         -- Defines the text that is prepended to all debug
//            output, typically the name of your application.
//            If you don't define BUGS_B_GONE_PREFIX yourself,
//            it defaults to "Bugs-B-Gone".
//
// Meet your slaves, oh master:
//
//     LOG((args))
//
//         -- Outputs the message given by "args". The argument
//            takes the variable argument form of a printf()
//            call and must be wrapped inside a second set of
//            parenthesis; for example: LOG(("Hi %s", "there!"))
//
//     ASSERT(args)
//
//         -- Outputs the name and line of the current source
//            code file if the condition given by "args" (which
//            can be any normal C or C++ expression) evaluates 
//            to false, and halts execution of the program.
//
//     CHECKPOINT
//
//         -- Outputs the name and line of the source code 
//            file that contains this checkpoint.
//
//     ENTER("function")
//
//         -- Outputs a string that indicates that the code 
//            has just entered a particular function.
//
//     LEAVE("function")
//
//         -- Outputs a string that indicates that the code 
//            is about to leave a particular function.
//
//     DEFINE(declarations)
//
//         -- Defines constants, variables, functions, or
//            what-have-you. These things are only available
//            when the BUGS_B_GONE symbol is defined.
//
//     EXEC(statements)
//
//         -- Executes a section of code given by "statements",
//            for example: EXEC(return -1)
//
// You don't need to separate these macros from other statements 
// with a semicolon.

#ifndef BUGS_B_GONE_H
#define BUGS_B_GONE_H

#ifndef BUGS_B_GONE_PREFIX
#define BUGS_B_GONE_PREFIX "Bugs-B-Gone"
#endif

#ifdef BUGS_B_GONE

#include <stdio.h>	// printf

#define LOG(args) {printf("%s ", BUGS_B_GONE_PREFIX); printf args; printf("\n");}
#define ASSERT(args) {if (!(args)) {LOG(("Assertion of \"%s\" failed at line %d of %s",#args,__LINE__,__FILE__)); abort();} }
#define CHECKPOINT {LOG(("Checkpoint at line %d of %s",__LINE__,__FILE__));}
#define ENTER(args) {LOG(("Entering \"%s\"...", args));}
#define LEAVE(args) {LOG(("Leaving \"%s\"...", args));}
#define DEFINE(args) args
#define EXEC(args) {args;}

#else // BUGS_B_GONE

#define LOG(args)
#define ASSERT(args)  
#define CHECKPOINT  
#define ENTER(args)
#define LEAVE(args)
#define DEFINE(args)
#define EXEC(args)

#endif // BUGS_B_GONE

#endif // BUGS_B_GONE_H
