// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <be/app/Roster.h>
#include <be/interface/Box.h>
#include <be/interface/Menu.h>
#include <be/interface/ScrollView.h>
#include <be/interface/Screen.h>
#include <be/storage/Entry.h>
#include <be/storage/Path.h>

#include <posix/ctype.h>
#include <posix/stdlib.h>
#include <posix/string.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "TrackerGrep.h"

#include "Grepper.h"
#include "GrepWindow.h"
#include "History.h"
#include "ResultItem.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

/** The messages that its views send to the window. */
enum
{
	 /** Start the grepping. */
	 MSG_START_SEARCH = 1000,
	
	 /** Cancel the grepping. */
	 MSG_CANCEL_SEARCH,
	
	 /** The user changed the "Follow symbolic links" setting. */
	 MSG_FOLLOW_LINKS_TOGGLED,
	 
	 /** The user changed the "Look in sub-directories" setting. */
	 MSG_SUBDIRECTORIES_TOGGLED,

	 /** The user changed the "Do not escape search text" setting. */
	 MSG_ESCAPE_TEXT_TOGGLED,

	 /** The user changed the "Case sensitive" setting. */
	 MSG_CASE_SENSITIVE_TOGGLED,
	
	 /** The value of the "Show contents" check box has changed. */
	 MSG_SHOWCONTENTS_TOGGLED,
	 
	 /** The user double-clicked on a result line. */
	 MSG_INVOKE_ITEM,
	 
	 /** The user changed the search pattern. */
	 MSG_SEARCH_TEXT_CHANGED,
};

/** The name of the settings file. */
const char* PREFS_FILE = "/boot/home/config/settings/TrackerGrepSettings";

/** The name of the history file. */
const char* HISTORY_FILE = "/boot/home/config/settings/TrackerGrepHistory";

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

GrepWindow::GrepWindow(entry_ref directory, BMessage* message)
	: BWindow(
		BRect(0, 0, 1, 1),
		"Tracker Grep " VERSION_STRING,
		B_DOCUMENT_WINDOW,
		0)
{
	// Initialize attributes.
	searchPattern = 0;
	busy = false;

	// Set the window's title to the directory we are invoked on.
	BEntry entry(&directory, true);
	if (entry.InitCheck() == B_OK)
	{
		BPath path;
		if (entry.GetPath(&path) == B_OK)
		{
			SetTitle(path.Path());
		}
	}
	
	// Create the grepper and history objects and connect them to us.
	grepper = new Grepper(directory, message);
	grepper->SetTarget(this);
	history = new History(this, 20, HISTORY_FILE);

	// -----------------------------------------------------------------------

	BMenu* menu = new BMenu("Options");

	followLinksItem = new BMenuItem(
		"Follow symbolic links",
		new BMessage(MSG_FOLLOW_LINKS_TOGGLED));

	menu->AddItem(followLinksItem);

	subDirsItem = new BMenuItem(
		"Look in sub-directories",
		new BMessage(MSG_SUBDIRECTORIES_TOGGLED));

	menu->AddItem(subDirsItem);

	menu->AddSeparatorItem();

	escapeTextItem = new BMenuItem(
		"Do not escape search text",
		new BMessage(MSG_ESCAPE_TEXT_TOGGLED));

	menu->AddItem(escapeTextItem);

	caseSensitiveItem = new BMenuItem(
		"Case sensitive",
		new BMessage(MSG_CASE_SENSITIVE_TOGGLED));

	menu->AddItem(caseSensitiveItem);

	optionsMenu = new BMenuField(
		BRect(0, 0, 80, 17),
		"optionsMenu",
		0,
		menu,
		false,
		B_FOLLOW_LEFT | B_FOLLOW_TOP);

	// This doesn't really seem to return a useful size for BMenuFields so
	// we pass a BRect with a fixed size into its constructor. But we also
	// set the "fixedSize" argument to "false", so the width of the menu
	// field will be adjusted to the name of the BMenu (although this doesn't
	// change the width of the frame until the menu field is displayed, so
	// we won't be able to use the "real" width in our layout calculations).
	optionsMenu->ResizeToPreferred();

	// -----------------------------------------------------------------------

	// Create the search pattern's entry field. This view does not send a 
	// message when <Enter> is pressed, because the "Search/Cancel" button
	// already does this and we don't want to send the same message twice.
	searchText = new BTextControl(
		BRect(0, 0, 0, 1),
		"SearchText",
		0,                             // No label.
		0,                             // No initial text.
		0,                             // No message.
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
		B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE); 
	
	searchText->TextView()->SetMaxBytes(1000); 
	searchText->ResizeToPreferred();   // Because it doesn't have a label.

	searchText->SetModificationMessage(
		new BMessage(MSG_SEARCH_TEXT_CHANGED));

	// Create the history pop-up menu.
	historyMenu = new BMenuField(
		BRect(0, 0, 17, 17),
		"historyMenu",
		0,
		new BMenu(""),
		true,
		B_FOLLOW_RIGHT | B_FOLLOW_TOP);

	historyMenu->ResizeToPreferred();

	// Create the "Search/Cancel" button
	button = new BButton(
		BRect(0, 1, 80, 1),
		"Button",
		"Search",
		new BMessage(MSG_START_SEARCH),
		B_FOLLOW_RIGHT);
	
	button->MakeDefault(true);
	button->ResizeToPreferred();
	button->SetEnabled(false);
	
	// Create the "show contents" check box.
	showContents = new BCheckBox(
		BRect(0, 0, 1, 1),
		"ShowContents",
		"Show contents",
		new BMessage(MSG_SHOWCONTENTS_TOGGLED),
		B_FOLLOW_LEFT);
	
	showContents->SetValue(B_CONTROL_ON);
	showContents->ResizeToPreferred();
	
	// Create the search results display.
	searchResults = new BOutlineListView(
		BRect(0, 0, 40, 80),
		"SearchResults", 
		B_SINGLE_SELECTION_LIST,
		B_FOLLOW_ALL_SIDES,
		B_WILL_DRAW);
		
	searchResults->SetInvocationMessage(new BMessage(MSG_INVOKE_ITEM));
	searchResults->ResizeToPreferred();
	
	// Put scroll bars around the search results.  
	BScrollView* scroller = new BScrollView(
		"ScrollSearchResults", 
		searchResults, 
		B_FOLLOW_ALL_SIDES, 
		B_FULL_UPDATE_ON_RESIZE, 
		true,                        // Horizontal scrollbar.
		true,                        // Vertical scrollbar.
		B_NO_BORDER);

	scroller->ResizeToPreferred();

	// Create the background box.
	BBox* background = new BBox(
		BRect(0, 0, 1, 1),
		B_EMPTY_STRING, 
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 
		B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE_JUMP, 
		B_PLAIN_BORDER); 

	// -----------------------------------------------------------------------

	// Now all views have been created, we can 
	// calculate how large the window will be.
	float width = 8 + showContents->Frame().Width()
		+ 8 + optionsMenu->Frame().Width()
		+ 8 + button->Frame().Width() + 8;

	float height = 8 + searchText->Frame().Height() + 8
		+ button->Frame().Height() + 8 + scroller->Frame().Height();

	float backgroundHeight = 8 + searchText->Frame().Height()
		+ 8 + button->Frame().Height() + 8;

	// Resize the window to its final size.
	ResizeTo(width, height);  

	// Add the background box.	
	AddChild(background);
	background->ResizeTo(width,	backgroundHeight);

	// Add the search text entry field to the background box.
	background->AddChild(searchText);
	searchText->MoveTo(8, 8);
	searchText->ResizeBy(width - historyMenu->Frame().Width() - 24, 0);
	
	// Must come after AddChild().
	searchText->MakeFocus(true);          

	// Add the history pop-up menu to the background box.
	background->AddChild(historyMenu);
	historyMenu->MoveTo(
		width - 8 - historyMenu->Frame().Width(),
		8 + (searchText->Frame().Height() - historyMenu->Frame().Height())/2);

	// Add the "show contents" checkbox to the background box.
	background->AddChild(showContents);
	showContents->MoveTo(
		8,
		8 + searchText->Frame().Height() + 8
			+ (button->Frame().Height() - showContents->Frame().Height())/2);

	// Add the menu field to the background box.
	background->AddChild(optionsMenu);
	optionsMenu->MoveTo(
		8 + showContents->Frame().Width() + 8,
		8 + searchText->Frame().Height() + 8
			+ (button->Frame().Height() - optionsMenu->Frame().Height())/2);

	// Add the "Search/Cancel button" to the background box.
	background->AddChild(button);
	button->MoveTo(
		width - button->Frame().Width() - 8,
		8 + searchText->Frame().Height() + 8);

	// Add the search results scroller.
	AddChild(scroller);
	scroller->MoveTo(0, backgroundHeight + 1);
	scroller->ResizeTo(width + 1, height - backgroundHeight);

	// -----------------------------------------------------------------------
	
	// Center the window on the screen.
	BRect screenRect = BScreen(this).Frame();
	MoveTo(
		(screenRect.Width() - width) / 2,
		(screenRect.Height() - height) / 2);

	// Adjust size limits of the window.  
	SetSizeLimits(width, 10000, height, 10000);

	// Adjust the new window's settings to the last saved preferences.
	LoadPrefs();
	Show(); 	
}

// ===========================================================================
// Destructor
// ===========================================================================

GrepWindow::~GrepWindow(void)
{
	delete grepper;

	// The window already destroys the BViews.
}

// ===========================================================================
// FrameResized
// ===========================================================================

void 
GrepWindow::FrameResized(float width, float height)
{
	super::FrameResized(width, height);
	SavePrefs();
}

// ===========================================================================
// FrameMoved
// ===========================================================================

void
GrepWindow::FrameMoved(BPoint origin)
{
	super::FrameMoved(origin);
	SavePrefs();
}

// ===========================================================================
// MenusBeginning
// ===========================================================================

void
GrepWindow::MenusBeginning(void)
{
	history->FillMenu(historyMenu->Menu());
	super::MenusBeginning();
}

// ===========================================================================
// MenusEnded
// ===========================================================================

void
GrepWindow::MenusEnded(void)
{
	for (int32 t = historyMenu->Menu()->CountItems(); t > 0; --t)
	{
		delete historyMenu->Menu()->RemoveItem(t - 1);
	}
	super::MenusEnded();
}

// ===========================================================================
// MessageReceived
// ===========================================================================

void
GrepWindow::MessageReceived(BMessage* message)
{
	// Figure out what kind of message was received.
	switch (message->what)
	{
		// We need to start a new search. ------------------------------------
		case MSG_START_SEARCH:
		{
			// Quit the current grepper thread (if there is one).
			grepper->Stop();
			
			// Clear the search results view.
			searchResults->MakeEmpty();
			
			// Only start the search if the user actually entered some text.
			if (strcmp(searchText->Text(), "") != 0)
			{
				grepper->SetPattern(searchText->Text());
				history->Add(searchText->Text());
				if (grepper->Start() == B_OK)
				{
					busy = true;

					// From now on, we don't want to be notified when the
					// search pattern changes, because we will be putting
					// the names of the files we're looking at into it.
					searchText->SetModificationMessage(0);
	
					historyMenu->SetEnabled(false);
					optionsMenu->SetEnabled(false);
					searchText->SetEnabled(false);
					button->MakeFocus(true);
					button->SetLabel("Cancel");
					button->SetMessage(new BMessage(MSG_CANCEL_SEARCH));
					
					// Remember the search pattern.
					searchPattern = strdup(searchText->Text());
				}
			}
			break;
		}
		
		// We need to cancel the search. -------------------------------------
		case MSG_CANCEL_SEARCH:
		{
			grepper->Stop();
		
			// Fall through.
		}
		
		// The grepper object has finished searching. ------------------------
		case MSG_SEARCH_FINISHED:
		{
			historyMenu->SetEnabled(true);
			optionsMenu->SetEnabled(true);
			searchText->SetEnabled(true);
			searchText->MakeFocus(true);
			button->SetLabel("Search");
			button->SetMessage(new BMessage(MSG_START_SEARCH));
			
			// Put the search pattern back into the text control.
			searchText->SetText(searchPattern);
			searchText->TextView()->SelectAll();

			// From now on, we want to be notified when
			// the user changes the search pattern.
			searchText->SetModificationMessage(
				new BMessage(MSG_SEARCH_TEXT_CHANGED));

			busy = false;
			break;
		}
		
		// The grepper object sends us a file name. --------------------------
		case MSG_REPORT_FILE_NAME:
		{
			// The check for busy is necessary because there is a race
			// condition when grep is running and the user presses the
			// cancel button. The MSG_CANCEL_SEARCH message is received
			// before the last MSG_REPORT_FILE_NAME message comes in.
			// Without the check, the search pattern that was put back
			// after the cancel message was received will be overwritten
			// by the name of the file that was last reported.
			if (busy)
			{
				searchText->SetText(message->FindString("filename"));
			}
			break;
		}
		
		// The grepper object sends us the output of a grep operation. -------
		case MSG_REPORT_RESULT:
		{
			const char *buf;
			entry_ref ref;
			BListItem *item = 0;

			if (busy && (message->FindRef("ref", &ref) == B_OK))
			{
				// Add the top-level item for this file.
				item = new ResultItem(&ref);
				searchResults->AddItem(item);

				// And expand it if necessary.			
				item->SetExpanded(showContents->Value() == B_CONTROL_ON);

				// Find all the matching lines for this file.				
				type_code type;
				int32 count;
				message->GetInfo("text", &type, &count);
				while (message->FindString("text", --count, &buf) == B_OK)
				{
					// Replace all non-printable characters by spaces.				
					uchar* result = (uchar*) strdup(buf);
					int32 len = strlen((char*) result);
					for (int32 pos = 0; pos < len; ++pos)
					{
						if (!(result[pos] & 0x80))
						{
							if (iscntrl(result[pos]))
							{
								result[pos] = ' ';
							}
						}
					}

					// Add them under the top-level item.
					searchResults->AddUnder(
						new BStringItem((const char*) result), item);
					free(result);
				}
			}
			break;
		}
		
		// The grepper object lets us know a grep operation failed. ----------
		case MSG_REPORT_ERROR:
		{
			const char *buf;
			if (busy && (message->FindString("error", &buf) == B_OK))
			{
				searchResults->AddItem(new BStringItem(buf));
			}
			break;
		}

		// The user changed the "follow links" setting. ----------------------
		case MSG_FOLLOW_LINKS_TOGGLED:
		{
			grepper->SetRecurseLinks(!followLinksItem->IsMarked());
			followLinksItem->SetMarked(!followLinksItem->IsMarked());

			SavePrefs();
			break;
		}
		
		// The user changed the "subdirectories" setting. --------------------
		case MSG_SUBDIRECTORIES_TOGGLED:
		{
			grepper->SetRecurseDirs(!subDirsItem->IsMarked());
			subDirsItem->SetMarked(!subDirsItem->IsMarked());
			
			SavePrefs();
			break;
		}

		// The user changed the "escape text" setting. -----------------------
		case MSG_ESCAPE_TEXT_TOGGLED:
		{
			grepper->SetEscapePattern(escapeTextItem->IsMarked());
			escapeTextItem->SetMarked(!escapeTextItem->IsMarked());
			
			SavePrefs();
			break;
		}
		
		// The user changed the "case sensitive" setting. --------------------
		case MSG_CASE_SENSITIVE_TOGGLED:
		{
			grepper->SetCaseSensitive(!caseSensitiveItem->IsMarked());
			caseSensitiveItem->SetMarked(!caseSensitiveItem->IsMarked());
			
			SavePrefs();
			break;
		}
		
		// The user clicked on the "show contents" check box. ----------------
		case MSG_SHOWCONTENTS_TOGGLED:
		{
			// Loop through the list's items and expand or collapse them all.
			int32 numItems = searchResults->FullListCountItems();
			if (showContents->Value() == B_CONTROL_ON)
			{
				for (int32 index = 0; index < numItems; ++index)
				{
					searchResults->Expand(
						searchResults->FullListItemAt(index));
				}
			}
			else
			{
				for (int32 index = 0; index < numItems; ++index)
				{
					searchResults->Collapse(
						searchResults->FullListItemAt(index));
				}
			}
			SavePrefs();
			break;
		}
		
		// The user double-clicked on a result. ------------------------------
		case MSG_INVOKE_ITEM:
		{
			BListItem* item = searchResults->ItemAt(
				searchResults->CurrentSelection());
			
			if (item != 0)
			{
				// Get the top-most item.
				int32 level = item->OutlineLevel();
				while (level != 0)
				{
					item = searchResults->Superitem(item);
					level = item->OutlineLevel();
				}
			
				// And launch its entry_ref.
				ResultItem* entry = dynamic_cast<ResultItem*>(item);
				if (entry != 0)
				{
					entry_ref* ref = entry->GetRef();
					be_roster->Launch(ref);
				}
			}	
			break;
		}
	
		// The user changed the search pattern. ------------------------------
		case MSG_SEARCH_TEXT_CHANGED:
		{
			button->SetEnabled(strcmp(searchText->Text(), "") != 0);
			break;
		}

		// The user picked an item from the history. -------------------------
		case MSG_HISTORY_ITEM_SELECTED:
		{
			const char *buf;
			if (message->FindString("text", &buf) == B_OK)
			{
				searchText->SetText(buf);
			}
			break;
		}

		// Let our super class handle the message. ---------------------------
		default:
		{
			super::MessageReceived(message);
			break;
		}
	}
}

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ===========================================================================
// LoadPrefs
// ===========================================================================

void
GrepWindow::LoadPrefs(void)
{
	BFile file(PREFS_FILE, B_READ_ONLY);
	if (file.InitCheck() == B_OK)
	{
		Lock();

		if (file.Lock() == B_OK)
		{
			int32 value = 0;
			file.ReadAttr(
				"RecurseDirs",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
			
			subDirsItem->SetMarked(value == B_CONTROL_ON);
			grepper->SetRecurseDirs(value == B_CONTROL_ON);
		
			file.ReadAttr(
				"RecurseLinks",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
				
			followLinksItem->SetMarked(value == B_CONTROL_ON);
			grepper->SetRecurseLinks(value == B_CONTROL_ON);

			file.ReadAttr(
				"EscapeText",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
				
			escapeTextItem->SetMarked(value != B_CONTROL_ON);
			grepper->SetEscapePattern(value == B_CONTROL_ON);

			file.ReadAttr(
				"CaseSensitive",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
				
			caseSensitiveItem->SetMarked(value == B_CONTROL_ON);
			grepper->SetCaseSensitive(value == B_CONTROL_ON);
			
			file.ReadAttr(
				"ShowContents",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
				
			showContents->SetValue(value);
			
			BRect rect;
			file.ReadAttr(
				"WindowFrame",
				B_RECT_TYPE,
				0,
				&rect,
				sizeof(BRect));
				
			MoveTo(rect.left, rect.top);
			ResizeTo(rect.Width(), rect.Height());
			
			file.Unlock();
		}

		Unlock();
	}
}

// ===========================================================================
// SavePrefs
// ===========================================================================

void
GrepWindow::SavePrefs(void)
{
	BFile file(PREFS_FILE, B_CREATE_FILE | B_ERASE_FILE | B_WRITE_ONLY);
	if (file.InitCheck() == B_OK)
	{
		// Lock the window and the file. (Note that the window is
		// usually already locked, because SavePrefs is called from
		// MessageReceived()).
		Lock();
		
		if (file.Lock() == B_OK)
		{
			int32 value = 2;
			file.WriteAttr(
				"Version",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
		
			value = B_CONTROL_OFF;
			if (subDirsItem->IsMarked())
			{
				value = B_CONTROL_ON;
			}
			
			file.WriteAttr(
				"RecurseDirs",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));

			value = B_CONTROL_OFF;
			if (followLinksItem->IsMarked())
			{
				value = B_CONTROL_ON;
			}
						
			file.WriteAttr(
				"RecurseLinks",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));

			value = B_CONTROL_OFF;
			if (!escapeTextItem->IsMarked())
			{
				value = B_CONTROL_ON;
			}
						
			file.WriteAttr(
				"EscapeText",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));

			value = B_CONTROL_OFF;
			if (caseSensitiveItem->IsMarked())
			{
				value = B_CONTROL_ON;
			}
						
			file.WriteAttr(
				"CaseSensitive",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
			
			value = showContents->Value();
			file.WriteAttr(
				"ShowContents",
				B_INT32_TYPE,
				0,
				&value,
				sizeof(int32));
			
			BRect rect = Frame();
			file.WriteAttr(
				"WindowFrame",
				B_RECT_TYPE,
				0,
				&rect,
				sizeof(BRect));
			
			// Write the file immediately to disk, and then
			// unlock everything.
			file.Sync();
			file.Unlock();
		}

		Unlock();
	}
}

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
