// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <be/storage/Entry.h>
#include <be/storage/NodeInfo.h>
#include <be/storage/Path.h>
#include <be/storage/StorageDefs.h>

#include <posix/stdio.h>
#include <posix/stdlib.h>
#include <posix/string.h>
#include <posix/unistd.h>
#include <posix/sys/stat.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "TrackerGrep.h"
#include "Grepper.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

Grepper::Grepper(entry_ref directory, BMessage* message) 
{
	// Initialize attributes.
	target = 0;
	recurseDirs = true;
	recurseLinks = false;
	escapePattern = true;
	caseSensitive = false;
	threadMustQuit = false;
	grepperThreadId = -1;
	directories = new BList(10);

	pattern = 0;
	SetPattern("");
	
	// Allocate a BDirectory object to remember the
	// initial working directory.
	initialDir = new BDirectory(&directory);
	initialDir->Rewind();

	// Make a copy of the message, since we don't own it.
	this->message = new BMessage(*message);
}

// ===========================================================================
// Destructor
// ===========================================================================

Grepper::~Grepper(void)
{
	// Wait for the thread to quit.
	Stop();
	
	// Clean up.
	delete initialDir;
	delete directories;
	delete message;
	free(pattern);
}

// ===========================================================================
// SetTarget
// ===========================================================================

void
Grepper::SetTarget(BLooper* target)
{
	this->target = target;
}

// ===========================================================================
// SetPattern
// ===========================================================================

void
Grepper::SetPattern(const char* pattern)
{
	free(this->pattern);
	
	// Do we need to escape the pattern?
	if (escapePattern)
	{
		// We will simply guess the size of the memory buffer
		// that we need. This should always be large enough.
		int32 size = 2 + strlen(pattern) * 3;
		this->pattern = (char*) malloc(size * sizeof(char));
		
		const char* pat_ptr = pattern;
		char* buf_ptr = this->pattern;
		char c;

		// Put double quotes around the pattern, so separate
		// words are considered to be part of a single string.
		*buf_ptr++ = '"';

		while (*pat_ptr != '\0')
		{
			c = *pat_ptr++;
			
			// Put a backslash in front of characters
			// that should be escaped. 
			if ((c == '.')  || (c == ',')
			||  (c == '[')  || (c == ']')
			||  (c == '?')  || (c == '*')
			||  (c == '+')  || (c == '-')
			||  (c == ':')  || (c == '^')
			||  (c == '\'') || (c == '"'))
			{
				*buf_ptr++ = '\\';
			}
			else if ((c == '\\') || (c == '$'))
			{
				// Some characters need to be escaped
				// with *three* backslashes in a row.
				*buf_ptr++ = '\\';
				*buf_ptr++ = '\\';
				*buf_ptr++ = '\\';
			}

			// Note: we do not have to escape the
			// { } ( ) < > and | characters.

			*buf_ptr++ = c;
		}

		// Finish the pattern string.
		*buf_ptr++ = '"';
		*buf_ptr = '\0';		
	}
	else
	{
		// No need to escape, so simply make a straight copy of the text.
		this->pattern = strdup(pattern);
	}
}

// ===========================================================================
// SetRecurseDirs
// ===========================================================================

void 
Grepper::SetRecurseDirs(bool flag)
{
	recurseDirs = flag;
}

// ===========================================================================
// SetRecurseLinks
// ===========================================================================

void 
Grepper::SetRecurseLinks(bool flag)
{
	recurseLinks = flag;
}

// ===========================================================================
// SetEscapePattern
// ===========================================================================

void 
Grepper::SetEscapePattern(bool flag)
{
	escapePattern = flag;
}

// ===========================================================================
// SetCaseSensitive
// ===========================================================================

void 
Grepper::SetCaseSensitive(bool flag)
{
	caseSensitive = flag;
}

// ===========================================================================
// Start
// ===========================================================================

status_t
Grepper::Start(void)
{
	// Add a *copy* of the initial directory to the list of subdirectories.
	// (We need to add a copy to preserve the original initial directory
	// object, since we'll delete the entire list and its contents in the
	// Stop() function.
	currentDir = new BDirectory(*initialDir);
	directories->AddItem(currentDir);
	currentRef = 0;
	
	// Start the new thread.
	threadMustQuit = false;

	grepperThreadId =
		spawn_thread(SpawnThread, "GrepperThread", B_NORMAL_PRIORITY, this); 
	                                
	return (resume_thread(grepperThreadId)); 
}

// ===========================================================================
// Stop
// ===========================================================================

void
Grepper::Stop(void)
{
	// Tell the thread to quit, and wait until it actually has.
	threadMustQuit = true;
	int32 exitValue;
	wait_for_thread(grepperThreadId, &exitValue);
	
	// Remove all BDirectory objects from the list and free them. (If the
	// thread terminated normally, then there is only one object: the copy
	// of the initial directory. But if the user aborted the search, there
	// may be more objects in the list.)
	for (int32 count = directories->CountItems(); count > 0; --count)
	{
		delete directories->RemoveItem((int32) 0);
	}
}

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ===========================================================================
// SpawnThread
// ===========================================================================

int32 
Grepper::SpawnThread(void* arg) 
{ 
	return ((Grepper*) arg)->GrepperThread();
}
   
// ===========================================================================
// GrepperThread
// ===========================================================================

int32 
Grepper::GrepperThread(void) 
{
	char fileName[MAX_LEN_REPORT_TEXT]; 
	char tempString[MAX_LEN_REPORT_TEXT];
	char command[B_PATH_NAME_LENGTH + 32];
	BMessage* msg;

	// Initialize the temporary file.  
	BPath tempFile;
	char tempFileName[B_FILE_NAME_LENGTH + 30];
	sprintf(tempFileName, "/boot/var/tmp/temp%ld", grepperThreadId);
	tempFile.SetTo(tempFileName);

	// The case insensitivity grep option.
	char* caseOption = "";
	if (!caseSensitive)
	{
		caseOption = "-i";
	}

	// Loop through all the available files.
	while (GetNextName(fileName) == B_OK)
	{
		// Are we told to quit? Then exit the loop
		// (this terminates the thread).
		if (threadMustQuit)
		{
			break;
		}  
	
		// Send the name of the file to the target.
		strcpy(tempString, fileName);
		msg = new BMessage(MSG_REPORT_FILE_NAME);
		msg->AddString("filename", tempString);
		target->PostMessage(msg);
		delete msg;
		msg = 0;
	
		// If the file's contents actually match the pattern, we will
		// send another message that also contains the entry_ref of
		// the file for further use, like launching the file when it's
		// double-clicked.
		msg = new BMessage(MSG_REPORT_RESULT);
		msg->AddString("filename", tempString);
		
		BEntry entry(fileName);
		entry_ref ref;
		entry.GetRef(&ref);
		msg->AddRef("ref", &ref);

		// Execute grep and redirect its output to a temporary file.    
		AdjustFunnyCharacters(fileName);
		sprintf(command,
			"/boot/beos/bin/grep -hn %s %s \"%s\" > \"%s\"",
			caseOption,
			pattern,
			fileName,
			tempFile.Path());
		int success = system(command);

		// If grep ran successfully, then open the temporary 
		// file and send its contents to the target.
		FILE* results = 0;
		if (((success == 0) || (success == 1))
		&&  (results = fopen(tempFile.Path(), "r")))
		{
			while (fgets(tempString, MAX_LEN_REPORT_TEXT, results) != 0)
			{
				msg->AddString("text", tempString);
			}
	
			// This file is finished.
			if (!threadMustQuit && msg->HasString("text"))
			{
				target->PostMessage(msg);
			}
			delete msg;
			
			fclose(results);
		}
		else
		{
			// There was a problem running grep, or opening
			// the temporary file.
			msg = new BMessage(MSG_REPORT_ERROR);
			msg->AddString("error", "There was a problem running grep.");
			target->PostMessage(msg);
			delete msg;      
		}
	}
	
	// We remove the temporary file after the entire search has finished,
	// to prevent a whole lot of flickering in the corresponding Tracker
	// window if it might be open (which is very unlikely, though).
	remove(tempFile.Path());
	
	// Tell the target that the search has ended, unless we were terminated.
	if (!threadMustQuit)
	{
		msg = new BMessage(MSG_SEARCH_FINISHED);
		target->PostMessage(msg);
		delete msg;
	}
	
	return 0;
}

// ===========================================================================
// GetNextName
// ===========================================================================

status_t 
Grepper::GetNextName(char* buffer)
{
	// We will loop until we have found the name of a valid file,
	// or until there are no more files left.
	while (true)
	{
		// Get the BEntry of the next name. ----------------------------------
	
		// Remove the previous string.
		*buffer = 0;                
		BEntry entry;
	
		// We are not inside a subdirectory.
		int32 dirDepth = directories->CountItems();
		if (dirDepth == 1)
		{ 
			// There are two ways to retrieve the file names. If the user
			// selected one or more files, then we must look at the message
			// that was passed into Grepper's constructor. If the user
			// didn't select any files, we will simply read all the entries
			// from the current working directory.
			
			// We have found an entry_ref in the message.
			entry_ref fileRef;
			if (message->FindRef("refs", currentRef, &fileRef) == B_OK)
			{
				// Go to next entry_ref in the message.
				++currentRef;
				
				// Convert the entry_ref into a BEntry.
				entry.SetTo(&fileRef, recurseLinks);
			}
			
			// If there were no files selected, we will grep
			// the entire directory.
			else if (currentRef == 0)
			{
				// Get next entry from the current working directory.  
				if (currentDir->GetNextEntry(&entry, recurseLinks) != B_OK)
				{
					return B_ERROR;
				}
			}
			
			// There are no more entry_refs in the message
			// (currentRef > 0).
			else
			{ 
				return B_ERROR;
			}
		}

		// We are inside a subdirectory.
		else
		{
			// Get next entry from the current working directory.  
			if (currentDir->GetNextEntry(&entry, recurseLinks) != B_OK)
			{
				// No more entries in this subdirectory, so return
				// to its parent directory.
				directories->RemoveItem(currentDir);
				delete currentDir;
				currentDir = (BDirectory*) directories->LastItem();
				
				// Immediately start a new iteration of the loop (to
				// read the next BEntry from the parent subdirectory).
				continue;
			}
		}

		// Examine entry. ----------------------------------------------------
		
		// Get stat structure to determine whether the entry
		// is a file or subdirectory.
		struct stat fileStat;
		if (entry.GetStat(&fileStat) == B_OK)
		{ 
			// The entry is a directory.
			if (S_ISDIR(fileStat.st_mode))
			{
				// Do we want to recurse subdirectories ?
				if (recurseDirs)
				{
					BDirectory* dir = new BDirectory(&entry);
					if (dir->InitCheck() == B_OK)
					{
						// Add new directory to the list and
						// make it the current one
						directories->AddItem(dir);  
						currentDir = dir;
					}
				}
			}
			
			// The entry is a file or a (non-traversed) symbolic link.
			else
			{
				// Because grep was meant to look at text files, we will
				// add only those files to the list that have the MIME
				// supertype "text" or "message". (Note that it is possible
				// for grep to look at binary files, but the output is
				// usually very ugly.) At this point the BEntry object of
				// a symbolic link must already have been traversed
				// because the MIME type of a symbolic link itself is 
				// "application/x-vnd.Be-symlink" (which will be skipped).
				
				// Get the MIME type of the file.
				BNode node(&entry);
				BNodeInfo nodeInfo(&node);
				char mimeTypeString[B_MIME_TYPE_LENGTH];
				
				if (nodeInfo.GetType(mimeTypeString) == B_OK)
				{
					// Only continue if the MIME supertype 
					// is "text" or "message".
					BMimeType mimeType(mimeTypeString);
					BMimeType superType;
					if (mimeType.GetSupertype(&superType) == B_OK)
					{
						if ((strcmp("text", superType.Type()) == 0) 
						||  (strcmp("message", superType.Type()) == 0))
						{							
							// Get the full path name and 
							// copy it into the buffer.
							BPath path;
							if (entry.GetPath(&path) == B_OK)
							{ 
								strcpy(buffer, path.Path());
								return B_OK;
							}
						} 
					}
				}
			}
		}
		
		// We did find a file or directory but it wasn't valid
		// for our purposes. So we'll just try again.
	}
}

// ===========================================================================
// AdjustFunnyCharacters
// ===========================================================================

void
Grepper::AdjustFunnyCharacters(char* buffer)
{
	char* copyOfBuffer = strdup(buffer);
	for (uint32 count = 0; count < strlen(copyOfBuffer); ++count)
	{
		if (copyOfBuffer[count] == '"')
		{
			*buffer = '\\';      
			++buffer;
			*buffer = '"';
		}
		else
		{
			*buffer = copyOfBuffer[count];
		}
		++buffer;
	}
	*buffer = '\0';
	free(copyOfBuffer);
}

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
