#ifndef TRACKERGREP_GREPPER_H
#define TRACKERGREP_GREPPER_H

// ***************************************************************************
// System Includes
// ***************************************************************************

#include <be/app/Looper.h>
#include <be/app/Message.h>
#include <be/storage/Directory.h>
#include <be/support/List.h>

// ***************************************************************************
// Project Includes
// ***************************************************************************

// Standard type and constant definitions.
#include "StyleGuideDefs.h"

// ***************************************************************************
// External Name Declarations
// ***************************************************************************

// ***************************************************************************
// Constant Definitions
// ***************************************************************************

/** The messages that the Grepper object sends to its target. */
enum
{
	/** Sends the name of the next file that will be grepped. */
	MSG_REPORT_FILE_NAME = 2000,
	
	/**
	 * Sends the output of a grep operation. This message is only
	 * sent when a certain file's contents match the search pattern. 
	 */
	MSG_REPORT_RESULT,

	/** Sent when the grep operation failed. */
	MSG_REPORT_ERROR,
	
	/** Indicates that the search has finished. */
	MSG_SEARCH_FINISHED,
};

/** The maximum length of the "text" string inside a report message. */
const int32 MAX_LEN_REPORT_TEXT = B_PATH_NAME_LENGTH;

// ***************************************************************************
// Type Definitions
// ***************************************************************************

// ***************************************************************************
// Class Definition
// ***************************************************************************

/**
 * The Grepper class is responsible for executing "grep" with a certain
 * search pattern on a certain number of files.
 *
 * If the user selected one or more files, then only these files will be
 * processed. If nothing was selected, then all the files from the current
 * working directory will be grepped. Recursive crawling of subdirectories
 * is optional. While grep is running, the Grepper object will send messages
 * with the results of the operation to a certain target BLooper.
 *
 * Typical usage is as follows:
 *
 *     1) Create the Grepper object.
 *     2) Connect it to a target BLooper (for instance a BWindow).
 *     3) Set the search pattern.
 *     4) Set the options.
 *     5) Call Start() to begin.
 *     6) Do something else while the grepper thread does its job. 
 *     7) The grepper thread will terminate automatically when all the
 *        files are processed, but you may call Stop() to abort it.
 *
 * @author Matthijs Hollemans
 * @version 2.0 (Sunday, 24 September 2000)
 * @since Tracker Grep 2.0.0
 */
class Grepper 
{
public:
	// Constant Definitions ==================================================

	// Type Definitions ======================================================

	// Attribute Definitions =================================================

	// Member Function Definitions ===========================================

	/** 
	 * Constructor. 
	 *
	 * @param directory The entry_ref of the current directory. 
	 *        This may never be null!
	 * @param message A standard B_REFS_RECEIVED message with the
	 *        "refs" array containing the entry_refs of the files
	 *        selected by the user. This may never be null!
	 */
	Grepper(entry_ref directory, BMessage* message);

	/** Automatically generated copy constructor. */
	// Grepper(const Grepper& old);

	/** Destructor. */
	virtual ~Grepper(void);

	/** Automatically generated assignment operator. */
	// Grepper& operator=(Grepper& right);

	/** Sets the target BLooper that the we will send our messages to. */
	void SetTarget(BLooper* target);
	
	/** Sets the search pattern. */
	void SetPattern(const char* pattern);
	
	/**
	 * Determines whether subdirectories should be crawled or not. 
	 * Defaults to true.
	 */
	void SetRecurseDirs(bool flag);
	
	/**
	 * Determines whether symbolic links should be traversed or not. 
	 * Defaults to false.
	 */
	void SetRecurseLinks(bool flag);

	/**
	 * Determines whether special characters in the search pattern
	 * should be escaped. Defaults to true.
	 */
	void SetEscapePattern(bool flag);

	/**
	 * Determines whether the search pattern is case sensitive.
	 * Defaults to false.
	 */
	void SetCaseSensitive(bool flag);
	
	/** 
	 * Starts a new grepper thread. 
	 *
	 * The new thread isn't started if the search pattern is empty.
	 * Any running grepper threads are not stopped by this function.
	 * You will have to call Stop() yourself first.
	 *
	 * @returns B_OK if everything went fine, or an error code
	 *          otherwise. See resume_thread() in The Be Book,
	 *          Kernel Kit, Threads, for all possible return
	 *          values.
	 */
	status_t Start(void);
	
	/**
	 * Stops (and terminates) the grepper thread. 
	 *
	 * Note that the thread will not send a MSG_SEARCH_FINISHED
	 * message in this case.
	 */
	void Stop(void); 

protected:
	// Constant Definitions ==================================================

	// Type Definitions ======================================================

	// Attribute Definitions =================================================

	// Member Function Definitions ===========================================

private:
	// Constant Definitions ==================================================

	// Type Definitions ======================================================

	// Attribute Definitions =================================================
	
	/** List that contains pointers to BDirectory objects. */
	BList* directories;
	
	/**
	 * The current working directory at the time Tracker Grep was invoked.
	 */
	BDirectory* initialDir;
	
	/** The current working directory. (Used for subdirectory crawling.) */
	BDirectory* currentDir;
	
	/**
	 * A B_REFS_RECEIVED message containing the files selected by the user.
	 */
	BMessage* message;
	
	/** Number of the current entry_ref in the B_REFS_RECEIVED message. */
	int32 currentRef;
	  
	/** Pointer to the target BLooper. */
	BLooper* target;
	
	/** The search pattern. */
	char* pattern;
	    
	/** True if we need to look into subdirectories, false if we don't. */
	bool recurseDirs;
	
	/** True if we need to look into symbolic links, false if we don't. */
	bool recurseLinks;

	/** True if the search pattern will be escaped, false if it's not. */
	bool escapePattern;
	
	/** True if the search pattern is case sensitive, false if it's not. */
	bool caseSensitive;
	
	/** Thread ID of the current grepper thread. */
	thread_id grepperThreadId;
	
	/** True if the grepper thread must quit, false if it may continue. */ 
	bool threadMustQuit;
  
	// Member Function Definitions ===========================================

	/** Stub that spawns the real grepper thread. */
	static int32 SpawnThread(void* arg);
	
	/** 
	 * The thread function that does the actual grepping. 
	 *
	 * @returns Always 0.
	 */
	int32 GrepperThread(void); 
	
	/**
	 * Returns the full path name of the next file that needs to be grepped. 
	 *
	 * @param buffer GetNextName() copies the path name into this buffer.
	 *        It must be allocated before it's passed in. The buffer must
	 *        be large enough to accommodate the name; B_PATH_NAME_LENGTH
	 *        is a 100% safe bet.
	 *
	 * @returns B_OK if a name was found, B_ERROR if there are no more names.
	 */
	status_t GetNextName(char* buffer);
	
	/**
	 * Changes all " characters inside a certain string into \" characters.
	 *
	 * The BeOS allows file names to contain all characters, except for /.
	 * Unfortunately, the shell is more restrictive and can't handle spaces 
	 * inside file names. Therefore, Tracker Grep surrounds all file names
	 * with double quotes (") to prevent the shell from misinterpreting
	 * spaces, brackets and the like. But this causes problems for file
	 * names that already contain double quotes themselves. The solution is
	 * to prefix (or "escape") such characters with a back slash.
	 *
	 * @param buffer The string that we will be modifying. The buffer must
	 *        be large enough to accomodate the changes. B_PATH_NAME_LENGTH
	 *        is usually safe, since the name usually contains only a few "s.
	 */
	void AdjustFunnyCharacters(char* buffer);
};

// ***************************************************************************
// Global Variable Declarations
// ***************************************************************************

// ***************************************************************************
// Inline Function Definitions
// ***************************************************************************

// ***************************************************************************
// Helper Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Function Declarations
// ***************************************************************************

#endif // TRACKERGREP_GREPPER_H
